#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


import ccxt
import pytest
from os import path
from shutil import copyfile
import json

from config import SIMULATOR_INITIAL_STARTUP_PORTFOLIO, SIMULATOR_INITIAL_STARTUP_PORTFOLIO_VALUE, \
    REAL_INITIAL_STARTUP_PORTFOLIO, REAL_INITIAL_STARTUP_PORTFOLIO_VALUE, WATCHED_MARKETS_INITIAL_STARTUP_VALUES, \
    REFERENCE_MARKET, SIMULATOR_CURRENT_PORTFOLIO, CONFIG_TRADING, CONFIG_ENABLED_PERSISTENCE, CONFIG_TRADER, \
    CONFIG_ENABLED_OPTION
from octobot_trading.exchanges import ExchangeManager
from trading.trader.previous_trading_state_manager import PreviousTradingStateManager
from tests.test_utils.config import load_test_config
from tests.test_utils.config import TEST_CONFIG_FOLDER
from trading.trader.portfolio import Portfolio


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

TESTS_STATE_SAVE_FOLDER = f"{TEST_CONFIG_FOLDER}/trading_states_tests/"


async def init_default():
    config = load_test_config()
    exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
    await exchange_manager.initialize()
    exchange_inst = exchange_manager.get_exchange()
    return config, exchange_inst


async def test_enabled():
    config, _ = await init_default()
    assert not PreviousTradingStateManager.enabled(config)
    config[CONFIG_TRADING][CONFIG_ENABLED_PERSISTENCE] = True
    assert PreviousTradingStateManager.enabled(config)


async def test_load_ok_file():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ok_ref.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert not state_manager.first_data

    assert state_manager.has_previous_state(exchange_inst)

    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_CURRENT_PORTFOLIO) == {
        "BTC": 100,
        "USD": 9999
    }
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO) == {
        "BTC": 10,
        "USD": 1000
    }
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO) == {
        "ZYX": 9999,
        "DAI": 1
    }
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO_VALUE) == 10
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO_VALUE) == 100
    assert state_manager.get_previous_state(exchange_inst, WATCHED_MARKETS_INITIAL_STARTUP_VALUES) == {
        "BTC": 1,
        "USDT": 0.00025075791580050704,
        "XLM": 10,
        "ETC": 10,
        "XRB": 10,
        "VEN": 10,
        "ONT": 10,
        "XRP": 10,
        "NEO": 10,
        "ADA": 10,
        "EUR": 10,
        "POWR": 10,
        "WAX": 10,
        "ICX": 10,
        "XVG": 10,
        "ETH": 10
    }
    assert state_manager.get_previous_state(exchange_inst, REFERENCE_MARKET) == "BTC"


async def test_load_ok_file_2():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ok_ref.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok2_log")
    assert not state_manager.first_data
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_CURRENT_PORTFOLIO) == {
        "BTC": 100,
        "USD": 9999
    }
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO) == {
        "BTC": 10,
        "USD": 1000
    }
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO) == {
        "ZYX": 9999,
        "DAI": 1
    }
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO_VALUE) == 10
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO_VALUE) == 100
    assert state_manager.get_previous_state(exchange_inst, WATCHED_MARKETS_INITIAL_STARTUP_VALUES) == {
        "BTC": 1,
        "USDT": 0.00025075791580050704,
        "XLM": 10,
        "ETC": 10,
        "XRB": 10,
        "VEN": 10,
        "ONT": 10,
        "XRP": 10,
        "NEO": 10,
        "ADA": 10,
        "EUR": 10,
        "POWR": 10,
        "WAX": 10,
        "ICX": 10,
        "XVG": 10,
        "ETH": 10
    }
    assert state_manager.get_previous_state(exchange_inst, REFERENCE_MARKET) == "BTC"


async def test_update_state():
    config, exchange_inst = await init_default()
    e_name = exchange_inst.get_name()

    copyfile(f"{TESTS_STATE_SAVE_FOLDER}ok_ref.json", f"{TESTS_STATE_SAVE_FOLDER}to_modify.json")
    state_manager = PreviousTradingStateManager({e_name: None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}to_modify.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    state_manager.update_previous_states(exchange_inst, simulated_initial_portfolio={
        "BTC": {
            Portfolio.AVAILABLE: 10,
            Portfolio.TOTAL: 200
        }
    })
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO) == {
        "BTC": 200
    }
    state_manager.update_previous_states(exchange_inst, real_initial_portfolio={
        "BTC": {
            Portfolio.AVAILABLE: 10,
            Portfolio.TOTAL: 250
        }
    })
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO) == {
        "BTC": 250
    }
    state_manager.update_previous_states(exchange_inst, simulated_initial_portfolio_value=10)
    assert state_manager.get_previous_state(exchange_inst, SIMULATOR_INITIAL_STARTUP_PORTFOLIO_VALUE) == 10
    state_manager.update_previous_states(exchange_inst, real_initial_portfolio_value=100)
    assert state_manager.get_previous_state(exchange_inst, REAL_INITIAL_STARTUP_PORTFOLIO_VALUE) == 100
    state_manager.update_previous_states(exchange_inst, watched_markets_initial_values="xyz")
    assert state_manager.get_previous_state(exchange_inst, WATCHED_MARKETS_INITIAL_STARTUP_VALUES) == "xyz"
    state_manager.update_previous_states(exchange_inst, reference_market="LOL")
    assert state_manager.get_previous_state(exchange_inst, REFERENCE_MARKET) == "LOL"

    with open(f"{TESTS_STATE_SAVE_FOLDER}to_modify.json") as f:
        file_content = json.loads(f.read())
    assert file_content[e_name][SIMULATOR_INITIAL_STARTUP_PORTFOLIO] == {"BTC": 200}
    assert file_content[e_name][REAL_INITIAL_STARTUP_PORTFOLIO] == {"BTC": 250}
    assert file_content[e_name][SIMULATOR_INITIAL_STARTUP_PORTFOLIO_VALUE] == 10
    assert file_content[e_name][REAL_INITIAL_STARTUP_PORTFOLIO_VALUE] == 100
    assert file_content[e_name][WATCHED_MARKETS_INITIAL_STARTUP_VALUES] == "xyz"
    assert file_content[e_name][REFERENCE_MARKET] == "LOL"

    state_manager.reset_trading_history()


async def test_reset_history():
    config, exchange_inst = await init_default()
    copyfile(f"{TESTS_STATE_SAVE_FOLDER}ok_ref.json", f"{TESTS_STATE_SAVE_FOLDER}to_reset.json")
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}to_reset.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert not state_manager.first_data
    state_manager.reset_trading_history()

    assert not path.exists(f"{TESTS_STATE_SAVE_FOLDER}to_reset.json")

    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}to_reset.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data
    assert state_manager.should_initialize_data()


async def test_load_ko_file_1():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_ref_invalid_json.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_2():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_ref_symbols_changed.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_3():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_ref_ref_market_changed.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_4():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_ref_missing_key.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_5():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_ref_missing_exchange.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_6():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}NO_EXISTS.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_7():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ok_ref.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ko_log")
    assert state_manager.first_data


async def test_load_ko_file_8():
    config, exchange_inst = await init_default()
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_missing_trader_s_data.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data


async def test_load_ko_file_9():
    config, exchange_inst = await init_default()
    config[CONFIG_TRADER][CONFIG_ENABLED_OPTION] = True
    state_manager = PreviousTradingStateManager({exchange_inst.get_name(): None}, False, config,
                                                save_file=f"{TESTS_STATE_SAVE_FOLDER}ko_missing_trader_data.json",
                                                log_file=f"{TESTS_STATE_SAVE_FOLDER}ok_log")
    assert state_manager.first_data



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from config import DEFAULT_REFERENCE_MARKET, CONFIG_TRADING, CONFIG_TRADER_REFERENCE_MARKET
from config.config import load_config
from octobot.tools import ConfigManager
from octobot_commons.symbol_util import split_symbol


def test_get_market_pair():
    config = load_config("tests/static/config.json")

    pair, inverted = ConfigManager.get_market_pair(config, config[CONFIG_TRADING][CONFIG_TRADER_REFERENCE_MARKET])
    assert pair == ""
    assert inverted is False

    pair, inverted = ConfigManager.get_market_pair(config, "")
    assert pair == ""
    assert inverted is False

    pair, inverted = ConfigManager.get_market_pair(config, "VEN")
    assert pair == "VEN/BTC"
    assert inverted is False

    pair, inverted = ConfigManager.get_market_pair(config, "USDT")
    assert pair == "BTC/USDT"
    assert inverted is True

    pair, inverted = ConfigManager.get_market_pair(config, "XBT")
    assert pair == ""
    assert inverted is False

    # now change config reference market
    config[CONFIG_TRADING][CONFIG_TRADER_REFERENCE_MARKET] = "USDT"

    pair, inverted = ConfigManager.get_market_pair(config, "BTC")
    assert pair == "BTC/USDT"
    assert inverted is False

    pair, inverted = ConfigManager.get_market_pair(config, "VEN")
    assert pair == ""
    assert inverted is False

    config[CONFIG_TRADING].pop(CONFIG_TRADER_REFERENCE_MARKET)

    # now use config/constants.py reference market
    pair, inverted = ConfigManager.get_market_pair(config, "ADA")
    assert pair == "ADA/BTC"
    assert split_symbol(pair)[1] == DEFAULT_REFERENCE_MARKET
    assert inverted is False

    config.pop(CONFIG_TRADING)
    pair, inverted = ConfigManager.get_market_pair(config, "ADA")
    assert pair == ""
    assert inverted is False


def test_get_reference_market():
    config = load_config("tests/static/config.json")
    assert ConfigManager.get_reference_market(config) == DEFAULT_REFERENCE_MARKET

    config[CONFIG_TRADING][CONFIG_TRADER_REFERENCE_MARKET] = "ETH"
    assert ConfigManager.get_reference_market(config) == "ETH"



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from config.config import load_config


def test_load_config():
    result = load_config("tests/static/config.json")
    assert "crypto-currencies" in result
    assert "services" in result
    assert "exchanges" in result
    assert "trading" in result
    assert "Bitcoin" in result["crypto-currencies"]


def test_load_config_without_file(caplog):
    load_config("tests/static/test_config.json", error=False)
    assert 'file opening failed' in caplog.text


def test_load_config_incorrect(caplog):
    load_config("tests/unit_tests/tools_tests/incorrect_config_file.txt", error=False)
    assert 'json decoding failed' in caplog.text



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from evaluator.evaluator_creator import EvaluatorCreator
from evaluator.Strategies import StrategiesEvaluator
from tests.test_utils.config import load_test_config
from octobot_commons.tentacles_management.advanced_manager import AdvancedManager
from config import TimeFrames, CONFIG_EVALUATORS_WILDCARD
from evaluator.evaluator import Evaluator
from evaluator.TA import TAEvaluator
from evaluator.Social import SocialEvaluator


def _get_tools():
    config = load_test_config()
    AdvancedManager.create_class_list(config)
    evaluator = Evaluator()
    evaluator.set_config(config)
    evaluator.set_symbol("BTC/USDT")
    evaluator.set_time_frame(TimeFrames.ONE_HOUR)
    return evaluator, config


def _assert_created_instances(instances_list, super_class, config):
    class_list = [instance.__class__ for instance in instances_list]
    for eval_class in AdvancedManager.create_advanced_evaluator_types_list(super_class, config):
        eval_instance = eval_class()
        eval_instance.set_config(config)
        if eval_instance.get_is_enabled():
            assert eval_class in class_list


def test_create_ta_eval_list():
    evaluator, config = _get_tools()
    ta_list = EvaluatorCreator.create_ta_eval_list(evaluator, CONFIG_EVALUATORS_WILDCARD)
    _assert_created_instances(ta_list, TAEvaluator, config)


def test_create_social_eval_list():
    evaluator, config = _get_tools()
    so_list = EvaluatorCreator.create_social_eval(config, evaluator.symbol, [], CONFIG_EVALUATORS_WILDCARD)
    _assert_created_instances(so_list, SocialEvaluator, config)


def test_create_social_not_tasked_list():
    evaluator, config = _get_tools()
    so_list = EvaluatorCreator.create_social_eval(config, evaluator.symbol, [], CONFIG_EVALUATORS_WILDCARD)
    not_tasked_so_list = EvaluatorCreator.create_social_not_tasked_list(so_list)
    for evalator in not_tasked_so_list:
        assert not evalator.is_to_be_independently_tasked


def test_create_strategies_eval_list():
    evaluator, config = _get_tools()
    strat_list = EvaluatorCreator.create_strategies_eval_list(config)
    _assert_created_instances(strat_list, StrategiesEvaluator, config)


# not tested for now
# def test_create_real_time_ta_evals():
#     evaluator, config = _get_tools()
#     ta_list = EvaluatorCreator.create_real_time_ta_evals(config, evaluator.exchange,
#                                                          evaluator.symbol, CONFIG_EVALUATORS_WILDCARD)
#     _assert_created_instances(ta_list, RealTimeEvaluator, config)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import ccxt
import copy
import pytest

from octobot_trading.exchanges import ExchangeManager
from config import *
from tests.test_utils.config import load_test_config
from trading.trader.order import *
from trading.trader.order_notifier import OrderNotifier
from trading.trader.trader import Trader
from trading.trader.trader_simulator import TraderSimulator
from trading.trader.portfolio import Portfolio


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestTrader:
    DEFAULT_SYMBOL = "BTC/USDT"

    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 2)
        await trader_inst.portfolio.initialize()
        return config, exchange_inst, trader_inst

    @staticmethod
    def stop(trader):
        trader.stop_order_manager()

    async def test_enabled(self):
        config, _, trader_inst = await self.init_default()
        self.stop(trader_inst)

        config[CONFIG_TRADER][CONFIG_ENABLED_OPTION] = True
        assert Trader.enabled(config)

        config[CONFIG_TRADER][CONFIG_ENABLED_OPTION] = False
        assert not Trader.enabled(config)

    async def test_get_risk(self):
        config, exchange_inst, trader_inst = await self.init_default()
        self.stop(trader_inst)

        config[CONFIG_TRADING][CONFIG_TRADER_RISK] = 0
        trader_1 = TraderSimulator(config, exchange_inst)
        assert trader_1.get_risk() == CONFIG_TRADER_RISK_MIN
        self.stop(trader_1)

        config[CONFIG_TRADING][CONFIG_TRADER_RISK] = 2
        trader_2 = TraderSimulator(config, exchange_inst)
        assert trader_2.get_risk() == CONFIG_TRADER_RISK_MAX
        self.stop(trader_2)

        config[CONFIG_TRADING][CONFIG_TRADER_RISK] = 0.5
        trader_2 = TraderSimulator(config, exchange_inst)
        assert trader_2.get_risk() == 0.5
        self.stop(trader_2)

    async def test_cancel_order(self):
        _, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       self.DEFAULT_SYMBOL,
                       70,
                       10,
                       70)

        assert market_buy not in trader_inst.get_open_orders()

        trader_inst.get_order_manager().add_order_to_list(market_buy)

        assert market_buy in trader_inst.get_open_orders()

        await trader_inst.cancel_order(market_buy)

        assert market_buy not in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_cancel_open_orders_default_symbol(self):
        config, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       self.DEFAULT_SYMBOL,
                       70,
                       10,
                       70)

        # Test sell order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        self.DEFAULT_SYMBOL,
                        70,
                        10,
                        70)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      self.DEFAULT_SYMBOL,
                      70,
                      10,
                      70)

        # create order notifier to prevent None call
        market_buy.order_notifier = OrderNotifier(config, market_buy)
        market_sell.order_notifier = OrderNotifier(config, market_sell)
        limit_buy.order_notifier = OrderNotifier(config, limit_buy)

        trader_inst.get_order_manager().add_order_to_list(market_buy)
        trader_inst.get_order_manager().add_order_to_list(market_sell)
        trader_inst.get_order_manager().add_order_to_list(limit_buy)

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell in trader_inst.get_open_orders()
        assert limit_buy in trader_inst.get_open_orders()

        await trader_inst.cancel_open_orders(self.DEFAULT_SYMBOL)

        assert market_buy not in trader_inst.get_open_orders()
        assert market_sell not in trader_inst.get_open_orders()
        assert limit_buy not in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_cancel_open_orders_multi_symbol(self):
        config, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/EUR",
                       70,
                       10,
                       70)

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "NANO/USDT",
                       70,
                       10,
                       70)

        # Test sell order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        self.DEFAULT_SYMBOL,
                        70,
                        10,
                        70)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      self.DEFAULT_SYMBOL,
                      70,
                      10,
                      70)

        # create order notifier to prevent None call
        market_buy.order_notifier = OrderNotifier(config, market_buy)
        market_sell.order_notifier = OrderNotifier(config, market_sell)
        limit_buy.order_notifier = OrderNotifier(config, limit_buy)
        limit_sell.order_notifier = OrderNotifier(config, limit_sell)

        trader_inst.get_order_manager().add_order_to_list(market_buy)
        trader_inst.get_order_manager().add_order_to_list(market_sell)
        trader_inst.get_order_manager().add_order_to_list(limit_buy)
        trader_inst.get_order_manager().add_order_to_list(limit_sell)

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell in trader_inst.get_open_orders()
        assert limit_buy in trader_inst.get_open_orders()
        assert limit_sell in trader_inst.get_open_orders()

        await trader_inst.cancel_open_orders(self.DEFAULT_SYMBOL)

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell not in trader_inst.get_open_orders()
        assert limit_buy not in trader_inst.get_open_orders()
        assert limit_sell in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_cancel_all_open_orders_with_currency(self):
        config, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/EUR",
                       70,
                       10,
                       70)

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "XRB/BTC",
                       70,
                       10,
                       70)

        # Test sell order
        market_sell = SellMarketOrder(trader_inst)
        market_sell.new(TraderOrderType.SELL_MARKET,
                        self.DEFAULT_SYMBOL,
                        70,
                        10,
                        70)

        # Test buy order
        limit_buy = BuyLimitOrder(trader_inst)
        limit_buy.new(TraderOrderType.BUY_LIMIT,
                      self.DEFAULT_SYMBOL,
                      70,
                      10,
                      70)

        # create order notifier to prevent None call
        market_buy.order_notifier = OrderNotifier(config, market_buy)
        market_sell.order_notifier = OrderNotifier(config, market_sell)
        limit_buy.order_notifier = OrderNotifier(config, limit_buy)
        limit_sell.order_notifier = OrderNotifier(config, limit_sell)

        trader_inst.get_order_manager().add_order_to_list(market_buy)
        trader_inst.get_order_manager().add_order_to_list(market_sell)
        trader_inst.get_order_manager().add_order_to_list(limit_buy)
        trader_inst.get_order_manager().add_order_to_list(limit_sell)

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell in trader_inst.get_open_orders()
        assert limit_buy in trader_inst.get_open_orders()
        assert limit_sell in trader_inst.get_open_orders()

        await trader_inst.cancel_all_open_orders_with_currency("XYZ")

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell in trader_inst.get_open_orders()
        assert limit_buy in trader_inst.get_open_orders()
        assert limit_sell in trader_inst.get_open_orders()

        await trader_inst.cancel_all_open_orders_with_currency("XRB")

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell in trader_inst.get_open_orders()
        assert limit_buy in trader_inst.get_open_orders()
        assert limit_sell not in trader_inst.get_open_orders()

        await trader_inst.cancel_all_open_orders_with_currency("USDT")

        assert market_buy in trader_inst.get_open_orders()
        assert market_sell not in trader_inst.get_open_orders()
        assert limit_buy not in trader_inst.get_open_orders()
        assert limit_sell not in trader_inst.get_open_orders()

        await trader_inst.cancel_all_open_orders_with_currency("BTC")

        assert market_buy not in trader_inst.get_open_orders()
        assert market_sell not in trader_inst.get_open_orders()
        assert limit_buy not in trader_inst.get_open_orders()
        assert limit_sell not in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_notify_order_close(self):
        config, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/EUR",
                       70,
                       10,
                       70)

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "NANO/USDT",
                       70,
                       10,
                       70)

        # Test stop loss order
        stop_loss = StopLossOrder(trader_inst)
        stop_loss.new(TraderOrderType.STOP_LOSS,
                      "BTC/USD",
                      60,
                      10,
                      60)

        # create order notifier to prevent None call
        market_buy.order_notifier = OrderNotifier(config, market_buy)
        limit_sell.order_notifier = OrderNotifier(config, limit_sell)
        stop_loss.order_notifier = OrderNotifier(config, stop_loss)

        trader_inst.get_order_manager().add_order_to_list(market_buy)
        trader_inst.get_order_manager().add_order_to_list(stop_loss)
        trader_inst.get_order_manager().add_order_to_list(limit_sell)

        await trader_inst.notify_order_close(limit_sell, True)
        await trader_inst.notify_order_close(market_buy, True)

        assert market_buy not in trader_inst.get_open_orders()
        assert limit_sell not in trader_inst.get_open_orders()
        assert stop_loss in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_notify_sell_limit_order_cancel(self):
        _, _, trader_inst = await self.init_default()
        initial_portfolio = copy.deepcopy(trader_inst.portfolio.portfolio)

        # Test buy order
        limit_buy = trader_inst.create_order_instance(order_type=TraderOrderType.BUY_LIMIT,
                                                      symbol="BQX/BTC",
                                                      current_price=4,
                                                      quantity=2,
                                                      price=4)

        await trader_inst.create_order(limit_buy, trader_inst.portfolio)

        await trader_inst.notify_order_close(limit_buy, True)

        assert limit_buy not in trader_inst.get_open_orders()

        assert initial_portfolio == trader_inst.portfolio.portfolio

        self.stop(trader_inst)

    async def test_notify_sell_limit_order_cancel_one_in_two(self):
        _, _, trader_inst = await self.init_default()
        initial_portfolio = copy.deepcopy(trader_inst.portfolio.portfolio)

        # Test buy order
        limit_buy = trader_inst.create_order_instance(order_type=TraderOrderType.BUY_LIMIT,
                                                      symbol="BQX/BTC",
                                                      current_price=4,
                                                      quantity=2,
                                                      price=4)

        await trader_inst.create_order(limit_buy, trader_inst.portfolio)

        # Test second buy order
        second_limit_buy = trader_inst.create_order_instance(order_type=TraderOrderType.BUY_LIMIT,
                                                             symbol="VEN/BTC",
                                                             current_price=1,
                                                             quantity=1.5,
                                                             price=1)

        await trader_inst.create_order(second_limit_buy, trader_inst.portfolio)

        # Cancel only 1st one
        await trader_inst.notify_order_close(limit_buy, True)

        assert limit_buy not in trader_inst.get_open_orders()
        assert second_limit_buy in trader_inst.get_open_orders()

        assert initial_portfolio != trader_inst.portfolio.portfolio
        assert trader_inst.portfolio.portfolio["BTC"][Portfolio.AVAILABLE] == 8.5
        assert trader_inst.portfolio.portfolio["BTC"][Portfolio.TOTAL] == 10

        self.stop(trader_inst)

    async def test_notify_sell_limit_order_fill(self):
        _, _, trader_inst = await self.init_default()
        initial_portfolio = copy.deepcopy(trader_inst.portfolio.portfolio)

        # Test buy order
        limit_buy = trader_inst.create_order_instance(order_type=TraderOrderType.BUY_LIMIT,
                                                      symbol="BQX/BTC",
                                                      current_price=0.1,
                                                      quantity=10,
                                                      price=0.1)

        await trader_inst.create_order(limit_buy, trader_inst.portfolio)

        limit_buy.filled_price = limit_buy.origin_price
        limit_buy.filled_quantity = limit_buy.origin_quantity

        await trader_inst.notify_order_close(limit_buy)

        assert limit_buy not in trader_inst.get_open_orders()

        assert initial_portfolio != trader_inst.portfolio.portfolio
        assert trader_inst.portfolio.portfolio["BTC"][Portfolio.AVAILABLE] == 9
        assert trader_inst.portfolio.portfolio["BTC"][Portfolio.TOTAL] == 9
        assert trader_inst.portfolio.portfolio["BQX"][Portfolio.AVAILABLE] == 10
        assert trader_inst.portfolio.portfolio["BQX"][Portfolio.TOTAL] == 10

        self.stop(trader_inst)

    async def test_notify_order_close_with_linked_orders(self):
        config, _, trader_inst = await self.init_default()

        # Test buy order
        market_buy = BuyMarketOrder(trader_inst)
        market_buy.new(TraderOrderType.BUY_MARKET,
                       "BTC/EUR",
                       70,
                       10,
                       70)

        # Test buy order
        limit_sell = SellLimitOrder(trader_inst)
        limit_sell.new(TraderOrderType.SELL_LIMIT,
                       "NANO/USDT",
                       70,
                       10,
                       70)

        # Test stop loss order
        stop_loss = StopLossOrder(trader_inst)
        stop_loss.new(TraderOrderType.STOP_LOSS,
                      "BTC/USD",
                      60,
                      10,
                      60)

        stop_loss.add_linked_order(limit_sell)
        limit_sell.add_linked_order(stop_loss)

        # create order notifier to prevent None call
        market_buy.order_notifier = OrderNotifier(config, market_buy)
        stop_loss.order_notifier = OrderNotifier(config, stop_loss)
        limit_sell.order_notifier = OrderNotifier(config, limit_sell)

        trader_inst.get_order_manager().add_order_to_list(market_buy)
        trader_inst.get_order_manager().add_order_to_list(stop_loss)
        trader_inst.get_order_manager().add_order_to_list(limit_sell)

        await trader_inst.notify_order_close(limit_sell)

        assert market_buy in trader_inst.get_open_orders()
        assert stop_loss not in trader_inst.get_open_orders()
        assert limit_sell not in trader_inst.get_open_orders()

        self.stop(trader_inst)

    async def test_parse_order_type(self):
        _, _, trader_inst = await self.init_default()

        ccxt_order_buy_market = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.MARKET
        }
        assert trader_inst.parse_order_type(ccxt_order_buy_market) == TraderOrderType.BUY_MARKET

        ccxt_order_buy_limit = {
            "side": TradeOrderSide.BUY,
            "type": TradeOrderType.LIMIT
        }
        assert trader_inst.parse_order_type(ccxt_order_buy_limit) == TraderOrderType.BUY_LIMIT

        ccxt_order_sell_market = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.MARKET
        }
        assert trader_inst.parse_order_type(ccxt_order_sell_market) == TraderOrderType.SELL_MARKET

        ccxt_order_sell_limit = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.LIMIT
        }
        assert trader_inst.parse_order_type(ccxt_order_sell_limit) == TraderOrderType.SELL_LIMIT

        self.stop(trader_inst)

    async def test_sell_all_currencies(self):
        _, _, trader_inst = await self.init_default()
        portfolio = trader_inst.portfolio.portfolio
        portfolio["ADA"] = {
            Portfolio.AVAILABLE: 1500,
            Portfolio.TOTAL: 1500
        }
        portfolio["USDT"] = {
            Portfolio.AVAILABLE: 1000,
            Portfolio.TOTAL: 1000
        }
        orders = await trader_inst.sell_all_currencies()
        # 1 order to sell ada, 1 order to buy btc (sell usdt), NO order for usd (not in config pairs)
        assert len(orders) == 2

        sell_ADA_order = orders[0]
        assert sell_ADA_order.symbol == "ADA/BTC"
        assert sell_ADA_order.order_type == TraderOrderType.SELL_MARKET
        assert sell_ADA_order.origin_quantity == 1500

        sell_USDT_order = orders[1]
        assert sell_USDT_order.symbol == "BTC/USDT"
        assert sell_USDT_order.order_type == TraderOrderType.BUY_MARKET
        assert round(sell_USDT_order.origin_quantity, 8) == round(1000/sell_USDT_order.origin_price, 8)

    async def test_sell_all(self):
        _, _, trader_inst = await self.init_default()
        portfolio = trader_inst.portfolio.portfolio
        portfolio["ADA"] = {
            Portfolio.AVAILABLE: 1500,
            Portfolio.TOTAL: 1500
        }
        portfolio["USDT"] = {
            Portfolio.AVAILABLE: 1000,
            Portfolio.TOTAL: 1000
        }
        orders = await trader_inst.sell_all("USDT")
        assert len(orders) == 1

        sell_USDT_order = orders[0]
        assert sell_USDT_order.symbol == "BTC/USDT"
        assert sell_USDT_order.order_type == TraderOrderType.BUY_MARKET
        assert round(sell_USDT_order.origin_quantity, 8) == round(1000/sell_USDT_order.origin_price, 8)

        orders = await trader_inst.sell_all("ADA")
        assert len(orders) == 1

        sell_ADA_order = orders[0]
        assert sell_ADA_order.symbol == "ADA/BTC"
        assert sell_ADA_order.order_type == TraderOrderType.SELL_MARKET
        assert sell_ADA_order.origin_quantity == 1500
        assert round(sell_USDT_order.origin_quantity, 8) == round(1000/sell_USDT_order.origin_price, 8)

        # currency not in portfolio
        orders = await trader_inst.sell_all("XBT")
        assert len(orders) == 0

        portfolio["XRB"] = {
            Portfolio.AVAILABLE: 0,
            Portfolio.TOTAL: 0
        }
        # currency in portfolio but with 0 quantity
        orders = await trader_inst.sell_all("XRB")
        assert len(orders) == 0

        # invalid currency
        orders = await trader_inst.sell_all("")
        assert len(orders) == 0

        portfolio["ICX"] = {
            Portfolio.AVAILABLE: 0.0000001,
            Portfolio.TOTAL: 0.0000001
        }
        # currency in portfolio but with close to 0 quantity
        orders = await trader_inst.sell_all("ICX")
        assert len(orders) == 0

    async def test_parse_exchange_order_to_trade_instance(self):
        _, exchange_inst, trader_inst = await self.init_default()

        order_to_test = Order(trader_inst)
        timestamp = time.time()

        exchange_order = {
            "status": OrderStatus.PARTIALLY_FILLED,
            "fee": 0.001,
            "price": 10.1444215411,
            "cost": 100.1444215411,
            "filled": 1.568415145687741563132,
            "timestamp": timestamp
        }

        trader_inst.parse_exchange_order_to_trade_instance(exchange_order, order_to_test)

        assert order_to_test.status == OrderStatus.PARTIALLY_FILLED
        assert order_to_test.filled_quantity == 1.568415145687741563132
        assert order_to_test.filled_price == 10.1444215411
        assert order_to_test.fee == 0.001
        assert order_to_test.total_cost == 100.1444215411
        assert order_to_test.executed_time == exchange_inst.get_uniform_timestamp(timestamp)

        self.stop(trader_inst)

    async def test_parse_exchange_order_to_order_instance(self):
        _, exchange_inst, trader_inst = await self.init_default()

        timestamp = time.time()

        exchange_order = {
            "side": TradeOrderSide.SELL,
            "type": TradeOrderType.LIMIT,
            "symbol": self.DEFAULT_SYMBOL,
            "amount": 1564.721672163722,
            "filled": 15.15467,
            "id": 1546541123,
            "status": OrderStatus.OPEN,
            "price": 10254.4515,
            "timestamp": timestamp
        }

        order_to_test = trader_inst.parse_exchange_order_to_order_instance(exchange_order)

        assert order_to_test.order_type == TraderOrderType.SELL_LIMIT
        assert order_to_test.status == OrderStatus.OPEN
        assert order_to_test.linked_to is None
        assert order_to_test.origin_stop_price is None
        assert order_to_test.origin_quantity == 1564.721672163722
        assert order_to_test.origin_price == 10254.4515
        assert order_to_test.filled_quantity == 1564.721672163722
        assert order_to_test.creation_time == exchange_inst.get_uniform_timestamp(timestamp)

        self.stop(trader_inst)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

from backtesting.backtesting_util import create_backtesting_config, create_backtesting_bot, start_backtesting_bot
from tests.test_utils.config import load_test_config


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

MULTIPLE_BACKTESTINGS_ITERATIONS = 10


async def test_simple_backtesting():
    config = create_backtesting_config(load_test_config(), ["ICX/BTC"])

    bot = create_backtesting_bot(config)
    await start_backtesting_bot(bot)


async def test_multiple_backtestings():
    config = create_backtesting_config(load_test_config(), ["ICX/BTC"])
    bot = create_backtesting_bot(config)
    previous_profitability, previous_market_profitability = await start_backtesting_bot(bot)
    for _ in range(MULTIPLE_BACKTESTINGS_ITERATIONS):
        config = create_backtesting_config(load_test_config(), ["ICX/BTC"])
        bot = create_backtesting_bot(config)
        current_profitability, current_market_profitability = await start_backtesting_bot(bot)
        assert previous_profitability == current_profitability
        assert previous_market_profitability == current_market_profitability


async def test_multiple_multi_symbol_backtestings():
    symbols = ["BTC/USDT", "NEO/BTC", "ICX/BTC", "VEN/BTC", "XRB/BTC"]
    config = create_backtesting_config(load_test_config(), symbols)
    bot = create_backtesting_bot(config)
    previous_profitability, previous_market_profitability = await start_backtesting_bot(bot)
    for _ in range(MULTIPLE_BACKTESTINGS_ITERATIONS):
        config = create_backtesting_config(load_test_config(), symbols)
        bot = create_backtesting_bot(config)
        current_profitability, current_market_profitability = await start_backtesting_bot(bot)
        assert previous_profitability == current_profitability
        assert previous_market_profitability == current_market_profitability



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from copy import deepcopy
import pytest

import ccxt

from tests.test_utils.config import load_test_config
from octobot_trading.exchanges import ExchangeManager
from octobot_trading.exchanges.websockets.binance_websocket import BinanceWebSocketClient


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestBinanceWebSocketClient:
    @staticmethod
    async def init_default():
        config = load_test_config()
        exchange_manager = ExchangeManager(config, ccxt.binance, True)
        await exchange_manager.initialize()
        binance_web_socket = BinanceWebSocketClient(config, exchange_manager)
        return config, binance_web_socket

    @staticmethod
    def price_message_header(stream, data):
        return {
            "stream": stream,
            "data": data
        }

    @staticmethod
    def _ticker_message(symbol):
        return TestBinanceWebSocketClient.price_message_header("@ticker", {
            "e": "24hrTicker",  # Event type
            "E": 123456789,  # Event time
            "s": symbol,  # Symbol
            "p": "0.0015",  # Price change
            "P": "250.00",  # Price change percent
            "w": "0.0018",  # Weighted average price
            "x": "0.0009",  # Previous day's close price
            "c": "0.0025",  # Current day's close price
            "Q": "10",  # Close trade's quantity
            "b": "0.0024",  # Best bid price
            "B": "10",  # Best bid quantity
            "a": "0.0026",  # Best ask price
            "A": "100",  # Best ask quantity
            "o": "0.0010",  # Open price
            "h": "0.0025",  # High price
            "l": "0.0010",  # Low price
            "v": "10000",  # Total traded base asset volume
            "q": "18",  # Total traded quote asset volume
            "O": 0,  # Statistics open time
            "C": 86400000,  # Statistics close time
            "F": 0,  # First trade ID
            "L": 18150,  # Last trade Id
            "n": 18151  # Total number of trades
        })

    @staticmethod
    def _kline_message(symbol, open_price, close_price, high_price, low_price, start_time, interval):
        return TestBinanceWebSocketClient.price_message_header("@kline", {
            "e": "kline",  # Event type
            "E": 123456789,  # Event time
            "s": symbol,  # Symbol
            "k": {
                "t": start_time,  # Kline start time
                "T": 123460000,  # Kline close time
                "s": "BNBBTC",  # Symbol
                "i": interval,  # Interval
                "f": 100,  # First trade ID
                "L": 200,  # Last trade ID
                "o": open_price,  # Open price
                "c": close_price,  # Close price
                "h": high_price,  # High price
                "l": low_price,  # Low price
                "v": "1000",  # Base asset volume
                "n": 100,  # Number of trades
                "x": False,  # Is this kline closed?
                "q": "1.0000",  # Quote asset volume
                "V": "500",  # Taker buy base asset volume
                "Q": "0.500",  # Taker buy quote asset volume
                "B": "123456"  # Ignore
            }
        })

    @staticmethod
    def _update_order_message(symbol, side, price, quantity, order_type, filled_qty, status):
        return {
            "e": "executionReport",  # Event type
            "E": 1499405658658,  # Event time
            "s": symbol,  # Symbol
            "c": "mUvoqJxFIILMdfAW5iGSOW",  # Client order ID
            "S": side,  # Side
            "o": order_type,  # Order type
            "f": "GTC",  # Time in force
            "q": quantity,  # Order quantity
            "p": price,  # Order price
            "P": "0.00000000",  # Stop price
            "F": "0.00000000",  # Iceberg quantity
            "g": -1,  # Ignore
            "C": "null",  # Original client order ID; This is the ID of the order being canceled
            "x": "NEW",  # Current execution type
            "X": status,  # Current order status
            "r": "NONE",  # Order reject reason; will be an error code.
            "i": 4293153,  # Order ID
            "l": "0.00000000",  # Last executed quantity
            "z": filled_qty,  # Cumulative filled quantity
            "L": "0.00000000",  # Last executed price
            "n": "0",  # Commission amount
            "N": None,  # Commission asset
            "T": 1499405658657,  # Transaction time
            "t": -1,  # Trade ID
            "I": 8641984,  # Ignore
            "w": True,  # Is the order working? Stops will have
            "m": False,  # Is this trade the maker side?
            "M": False  # Ignore
        }

    @staticmethod
    def _update_portfolio_message(symbols_data):
        msg = {
            "e": "outboundAccountInfo",
            "E": 1499405658849,
            "m": 0,
            "t": 0,
            "b": 0,  # Buyer commission rate (bips)
            "s": 0,  # Seller commission rate (bips)
            "T": True,  # Can trade?
            "W": True,  # Can withdraw?
            "D": True,  # Can deposit?
            "u": 1499405658848,  # Time of last account update
            "B": []
        }

        for symbol_data in symbols_data:
            msg["B"].append({
                "a": symbol_data["asset"],
                "f": symbol_data["free"],
                "l": symbol_data["locked"],
            })

        return msg

    async def test_update_portfolio(self):
        _, binance_web_socket = await self.init_default()

        binance_web_socket.get_personal_data().init_portfolio()

        origin_pf = deepcopy(binance_web_socket.get_personal_data().get_portfolio())

        # test with empty request
        binance_web_socket.user_callback(self._update_portfolio_message([]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert origin_pf == new_pf

        # test with not empty request
        binance_web_socket.user_callback(self._update_portfolio_message([{
            "asset": "BTC",
            "free": 0.5,
            "locked": 1,
        }]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert origin_pf != new_pf

        new_origin_pf = deepcopy(binance_web_socket.get_personal_data().get_portfolio())

        # test with empty request
        binance_web_socket.user_callback(self._update_portfolio_message([]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert new_origin_pf == new_pf

        # test with not empty request and not empty pf
        binance_web_socket.user_callback(self._update_portfolio_message([{
            "asset": "BTC",
            "free": 0.2,
            "locked": 15,
        }]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert new_origin_pf != new_pf
        new_origin_pf = deepcopy(binance_web_socket.get_personal_data().get_portfolio())

        # test with not empty request and diff symbol
        binance_web_socket.user_callback(self._update_portfolio_message([{
            "asset": "ETH",
            "free": 25.69,
            "locked": 31547,
        }]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert new_origin_pf["BTC"] == new_pf["BTC"]
        new_origin_pf = deepcopy(binance_web_socket.get_personal_data().get_portfolio())

        # test with not empty request and multiple symbol
        binance_web_socket.user_callback(self._update_portfolio_message([{
            "asset": "XRP",
            "free": 25978,
            "locked": 315047,
        }, {
            "asset": "LTC",
            "free": 25478.5877,
            "locked": 14875.1445,
        }, {
            "asset": "BCH",
            "free": 0.00015,
            "locked": 0.1055456,
        }]))
        new_pf = binance_web_socket.get_personal_data().get_portfolio()
        assert new_origin_pf["BTC"] == new_pf["BTC"]
        assert new_origin_pf["ETH"] == new_pf["ETH"]
        assert new_pf["XRP"]["free"] == 25978
        assert new_pf["LTC"]["used"] == 14875.1445
        assert new_pf["BCH"]["total"] == 0.00015 + 0.1055456

    async def test_set_ticker(self):
        _, binance_web_socket = await self.init_default()

        symbol = "BTCUSDT"

        symbol_data = binance_web_socket.get_symbol_data(symbol)
        symbol_data.update_symbol_ticker({})

        msg = self._ticker_message(symbol)
        binance_web_socket.all_currencies_prices_callback(msg)

        # assert symbol == symbol_data.symbol_ticker["symbol"]
        # assert symbol_data.symbol_ticker["info"] == msg["data"]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

import ccxt

from config import EvaluatorStates
from tests.test_utils.config import load_test_config
from octobot_trading.exchanges import ExchangeManager
from trading.modes import AbstractTradingModeCreator
from trading.trader.portfolio import Portfolio
from trading.trader.trader_simulator import TraderSimulator
from config import ExchangeConstantsMarketStatusColumns as Ecmsc


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestAbstractTradingModeCreator:
    @staticmethod
    async def _get_tools():
        config = load_test_config()
        symbol = "BTC/USDT"
        exchange_manager = ExchangeManager(config, ccxt.binance, is_simulated=True)
        await exchange_manager.initialize()
        exchange_inst = exchange_manager.get_exchange()
        trader_inst = TraderSimulator(config, exchange_inst, 0.3)
        await trader_inst.portfolio.initialize()
        trader_inst.stop_order_manager()
        trader_inst.portfolio.portfolio["SUB"] = {
            Portfolio.TOTAL: 0.000000000000000000005,
            Portfolio.AVAILABLE: 0.000000000000000000005
        }
        trader_inst.portfolio.portfolio["BNB"] = {
            Portfolio.TOTAL: 0.000000000000000000005,
            Portfolio.AVAILABLE: 0.000000000000000000005
        }
        trader_inst.portfolio.portfolio["USDT"] = {
            Portfolio.TOTAL: 2000,
            Portfolio.AVAILABLE: 2000
        }

        return config, exchange_inst, trader_inst, symbol

    async def test_can_create_order(self):
        _, exchange, trader, symbol = await self._get_tools()
        portfolio = trader.get_portfolio()
        # portfolio: "BTC": 10 "USD": 1000
        not_owned_symbol = "ETH/BTC"
        not_owned_market = "BTC/ETH"
        min_trigger_symbol = "SUB/BTC"
        min_trigger_market = "ADA/BNB"

        # order from neutral state => false
        assert not await AbstractTradingModeCreator(None).can_create_order(symbol, exchange,
                                                                           EvaluatorStates.NEUTRAL, portfolio)

        # sell order using a currency with 0 available
        assert not await AbstractTradingModeCreator(None).can_create_order(not_owned_symbol, exchange,
                                                                           EvaluatorStates.SHORT, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(not_owned_symbol, exchange,
                                                                           EvaluatorStates.VERY_SHORT, portfolio)

        # sell order using a currency with < min available
        assert not await AbstractTradingModeCreator(None).can_create_order(min_trigger_symbol, exchange,
                                                                           EvaluatorStates.SHORT, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(min_trigger_symbol, exchange,
                                                                           EvaluatorStates.VERY_SHORT, portfolio)

        # sell order using a currency with > min available
        assert await AbstractTradingModeCreator(None).can_create_order(not_owned_market, exchange,
                                                                       EvaluatorStates.SHORT, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(not_owned_market, exchange,
                                                                       EvaluatorStates.VERY_SHORT, portfolio)

        # buy order using a market with 0 available
        assert not await AbstractTradingModeCreator(None).can_create_order(not_owned_market, exchange,
                                                                           EvaluatorStates.LONG, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(not_owned_market, exchange,
                                                                           EvaluatorStates.VERY_LONG, portfolio)

        # buy order using a market with < min available
        assert not await AbstractTradingModeCreator(None).can_create_order(min_trigger_market, exchange,
                                                                           EvaluatorStates.LONG, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(min_trigger_market, exchange,
                                                                           EvaluatorStates.VERY_LONG, portfolio)

        # buy order using a market with > min available
        assert await AbstractTradingModeCreator(None).can_create_order(not_owned_symbol, exchange,
                                                                       EvaluatorStates.LONG, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(not_owned_symbol, exchange,
                                                                       EvaluatorStates.VERY_LONG, portfolio)

    async def test_can_create_order_unknown_symbols(self):
        _, exchange, trader, _ = await self._get_tools()
        portfolio = trader.get_portfolio()
        unknown_symbol = "VI?/BTC"
        unknown_market = "BTC/*s?"
        unknown_everything = "VI?/*s?"

        # buy order with unknown market
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_market, exchange,
                                                                           EvaluatorStates.LONG, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_market, exchange,
                                                                           EvaluatorStates.VERY_LONG, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(unknown_market, exchange,
                                                                       EvaluatorStates.SHORT, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(unknown_market, exchange,
                                                                       EvaluatorStates.VERY_SHORT, portfolio)

        # sell order with unknown symbol
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_symbol, exchange,
                                                                           EvaluatorStates.SHORT, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_symbol, exchange,
                                                                           EvaluatorStates.VERY_SHORT, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(unknown_symbol, exchange,
                                                                       EvaluatorStates.LONG, portfolio)
        assert await AbstractTradingModeCreator(None).can_create_order(unknown_symbol, exchange,
                                                                       EvaluatorStates.VERY_LONG, portfolio)

        # neutral state with unknown symbol, market and everything
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_symbol, exchange,
                                                                           EvaluatorStates.NEUTRAL, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_market, exchange,
                                                                           EvaluatorStates.NEUTRAL, portfolio)
        assert not await AbstractTradingModeCreator(None).can_create_order(unknown_everything, exchange,
                                                                           EvaluatorStates.NEUTRAL, portfolio)

    async def test_valid_create_new_order(self):
        _, exchange, trader, symbol = await self._get_tools()
        portfolio = trader.get_portfolio()
        order_creator = AbstractTradingModeCreator(None)

        # should raise NotImplementedError Exception
        try:
            order_creator.create_new_order(-1, symbol, exchange, trader, portfolio, EvaluatorStates.NEUTRAL)
            order_creator.create_new_order(1, symbol, exchange, trader, portfolio, EvaluatorStates.VERY_SHORT)
            order_creator.create_new_order(-0.65, symbol, exchange, trader, portfolio, EvaluatorStates.LONG)
            order_creator.create_new_order(-1, symbol, exchange, trader, portfolio, EvaluatorStates.VERY_LONG)
            assert False
        except NotImplementedError:
            assert True

    async def test_get_holdings_ratio(self):
        _, exchange, trader, symbol = await self._get_tools()
        new_portfolio = {
            "BTC": {
                Portfolio.TOTAL: 10
            },
            "USDT": {
                Portfolio.TOTAL: 1000
            }
        }
        order_creator = AbstractTradingModeCreator(None)
        trader.get_portfolio().portfolio = new_portfolio
        portfolio = trader.get_portfolio()
        ratio = await order_creator.get_holdings_ratio(trader, portfolio, "BTC")
        assert round(ratio, 8) == 0.9859353
        ratio = await order_creator.get_holdings_ratio(trader, portfolio, "USDT")
        assert round(ratio, 8) == 0.0140647

        new_portfolio.pop("USDT")
        ratio = await order_creator.get_holdings_ratio(trader, portfolio, "BTC")
        assert round(ratio, 8) == 1
        ratio = await order_creator.get_holdings_ratio(trader, portfolio, "USDT")
        assert round(ratio, 8) == 0
        ratio = await order_creator.get_holdings_ratio(trader, portfolio, "XYZ")
        assert round(ratio, 8) == 0

    async def test_get_number_of_traded_assets(self):
        _, exchange, trader, symbol = await self._get_tools()
        await trader.trades_manager.initialize()
        order_creator = AbstractTradingModeCreator(None)
        assert order_creator.get_number_of_traded_assets(trader) == 15

    async def test_adapt_price(self):
        # will use symbol market
        symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_PRICE.value: 4}}
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.0001) == 0.0001
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.00015) == 0.0001
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.005) == 0.005
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 1) == 1

        assert AbstractTradingModeCreator.adapt_price(symbol_market, 56.5128597145) == 56.5128
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 1251.0000014576121234854513) == 1251.0000

        # will use default (CURRENCY_DEFAULT_MAX_PRICE_DIGITS)
        symbol_market = {Ecmsc.PRECISION.value: {}}
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.0001) == 0.0001
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.00015) == 0.00014999
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 0.005) == 0.005
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 1) == 1.0000000000000000000000001
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 1) == 1

        assert AbstractTradingModeCreator.adapt_price(symbol_market, 56.5128597145) == 56.51285971
        assert AbstractTradingModeCreator.adapt_price(symbol_market, 1251.0000014576121234854513) == 1251.00000145

    async def test_get_additional_dusts_to_quantity_if_necessary(self):
        symbol_market = {Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1
            }
        }}

        current_symbol_holding = 5
        quantity = 3
        price = 1
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == quantity + 0

        current_symbol_holding = 5
        quantity = 4
        price = 1
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == quantity + 1

        current_symbol_holding = 5
        quantity = 4.5
        price = 1
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == quantity + 0.5

        symbol_market = {Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.005
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 0.00005
            }
        }}

        current_symbol_holding = 0.99000000001
        quantity = 0.9
        price = 0.5
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == quantity + 0

        current_symbol_holding = 0.99000000001
        quantity = 0.0215245845
        price = 0.5
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == quantity + 0

        current_symbol_holding = 0.99999999
        quantity = 0.99999
        price = 0.5
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == 0.99999999

        current_symbol_holding = 0.88
        quantity = 0.7055680057024826
        price = 0.0002
        assert AbstractTradingModeCreator.add_dusts_to_quantity_if_necessary(quantity,
                                                                             price,
                                                                             symbol_market,
                                                                             current_symbol_holding) == 0.88

    async def test_check_and_adapt_order_details_if_necessary(self):
        atmc = AbstractTradingModeCreator(None)

        symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: 1,
                    Ecmsc.LIMITS_COST_MAX.value: 200
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                    Ecmsc.LIMITS_PRICE_MAX.value: 50
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 8,
                Ecmsc.PRECISION_AMOUNT.value: 8
            }
        }

        invalid_cost_symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: None,
                    Ecmsc.LIMITS_COST_MAX.value: None
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                    Ecmsc.LIMITS_PRICE_MAX.value: 50
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 8,
                Ecmsc.PRECISION_AMOUNT.value: 8
            }
        }

        # correct min
        quantity = 0.5
        price = 2
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # correct max
        quantity = 100
        price = 2
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # correct
        quantity = 10
        price = 0.6
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # correct
        quantity = 3
        price = 49.9
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # invalid price > but valid cost
        quantity = 1
        price = 100
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # invalid price < but valid cost
        quantity = 10
        price = 0.1
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(quantity, price)]

        # invalid price > and invalid cost from exchange => use price => invalid
        quantity = 1
        price = 100
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

        # invalid price < and invalid cost from exchange => use price => invalid
        quantity = 1
        price = 0.1
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

        # invalid cost <
        quantity = 0.5
        price = 1
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

        # invalid cost >
        quantity = 10
        price = 49
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(1.83673469, 49),
                                                                                                   (4.08163265, 49),
                                                                                                   (4.08163265, 49)]

        # high cost but no max cost => valid
        quantity = 10
        price = 49
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(1.83673469, 49),
                                                                                                   (4.08163265, 49),
                                                                                                   (4.08163265, 49)]

        # invalid cost with invalid price >=
        quantity = 10
        price = 50
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(2, 50), (4, 50),
                                                                                                   (4, 50)]

        # invalid cost with invalid price > and invalid cost from exchange => use price => invalid
        quantity = 10
        price = 51
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, invalid_cost_symbol_market) == []

        # invalid amount >
        quantity = 200
        price = 5
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [(40.0, 5), (40.0, 5),
                                                                                                   (40.0, 5), (40.0, 5),
                                                                                                   (40.0, 5)]

        # invalid amount <
        quantity = 0.4
        price = 5
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

        symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                    Ecmsc.LIMITS_COST_MAX.value: 10
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                    Ecmsc.LIMITS_PRICE_MAX.value: 50
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 8,
                Ecmsc.PRECISION_AMOUNT.value: 8
            }
        }

        # correct quantity
        # to test _adapt_order_quantity_because_quantity
        quantity = 5000
        price = 0.001
        expected = [(100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001), (100.0, 0.001),
                    (100.0, 0.001), (100.0, 0.001)]
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == expected

        symbol_market_without_max = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: None,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                    Ecmsc.LIMITS_COST_MAX.value: None
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                    Ecmsc.LIMITS_PRICE_MAX.value: None
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 8,
                Ecmsc.PRECISION_AMOUNT.value: 8
            }
        }

        # high cost but no max cost => no split
        quantity = 10
        price = 49
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == [(10, 49)]

        # high quantity but no max quantity => no split
        quantity = 10000000
        price = 49
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == \
            [(10000000, 49)]

        # high price but no max price => no split
        quantity = 10
        price = 4900000
        assert atmc.check_and_adapt_order_details_if_necessary(quantity, price, symbol_market_without_max) == \
            [(10, 4900000)]

    async def test_get_pre_order_data(self):
        pass

    async def test_split_orders(self):
        atmc = AbstractTradingModeCreator(None)

        symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: 1,
                    Ecmsc.LIMITS_COST_MAX.value: 30
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.001,
                    Ecmsc.LIMITS_PRICE_MAX.value: 1000
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 1,
                Ecmsc.PRECISION_AMOUNT.value: 1
            }
        }

        symbol_market_without_max = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: None,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: 1,
                    Ecmsc.LIMITS_COST_MAX.value: None
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 1,
                    Ecmsc.LIMITS_PRICE_MAX.value: None
                },
            },
            Ecmsc.PRECISION.value: {
                Ecmsc.PRECISION_PRICE.value: 1,
                Ecmsc.PRECISION_AMOUNT.value: 1
            }
        }
        max_cost = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MAX.value]
        max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]

        # normal situation, split because of cost
        total_price = 100
        valid_quantity = 5
        price = 20
        assert atmc._split_orders(total_price, max_cost, valid_quantity,
                                  max_quantity, price, valid_quantity, symbol_market) \
            == [(0.5, 20), (1.5, 20), (1.5, 20), (1.5, 20)]

        # normal situation, split because of quantity
        total_price = 5.0255
        valid_quantity = 502.55
        price = 0.01
        assert atmc._split_orders(total_price, max_cost, valid_quantity,
                                  max_quantity, price, valid_quantity, symbol_market) \
            == [(2.5, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01)]

        # missing info situation, split because of cost
        max_quantity = None
        total_price = 100
        valid_quantity = 5
        price = 20
        assert atmc._split_orders(total_price, max_cost, valid_quantity,
                                  max_quantity, price, valid_quantity, symbol_market_without_max) \
            == [(0.5, 20), (1.5, 20), (1.5, 20), (1.5, 20)]

        # missing info situation, split because of quantity
        max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]
        max_cost = None
        total_price = 5.0255
        valid_quantity = 502.55
        price = 0.01
        assert atmc._split_orders(total_price, max_cost, valid_quantity,
                                  max_quantity, price, valid_quantity, symbol_market_without_max) \
            == [(2.5, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01), (100, 0.01)]

        # missing info situation, can't split
        max_quantity = None
        max_cost = None
        total_price = 5.0255
        valid_quantity = 502.55
        price = 0.01
        with pytest.raises(RuntimeError):
            assert atmc._split_orders(total_price, max_cost, valid_quantity,
                                      max_quantity, price, valid_quantity, symbol_market_without_max)

    async def test_adapt_quantity(self):
        # will use symbol market
        symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_AMOUNT.value: 4}}
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.0001) == 0.0001
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.00015) == 0.0001
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.005) == 0.005
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1) == 1.0000000000000000000000001
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1) == 1

        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 56.5128597145) == 56.5128
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1251.0000014576121234854513) == 1251.0000

        # will use default (0)
        symbol_market = {Ecmsc.PRECISION.value: {}}
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.0001) == 0
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.00015) == 0
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 0.005) == 0
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1) == 1.0000000000000000000000001
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1) == 1

        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 56.5128597145) == 56
        assert AbstractTradingModeCreator._adapt_quantity(symbol_market, 1251.0000014576121234854513) == 1251

    @staticmethod
    async def test_trunc_with_n_decimal_digits():
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(1.00000000001, 10) == 1
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(1.00000000001, 11) == 1.00000000001
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 3) == 578
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 4) == 578.0001
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 7) == 578.000145
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 9) == 578.000145
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 10) == 578.0001450001
        assert AbstractTradingModeCreator._trunc_with_n_decimal_digits(578.000145000156, 12) == 578.000145000156

    @staticmethod
    async def test_get_min_max_amounts():
        # normal values
        symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: None,
                    Ecmsc.LIMITS_COST_MAX.value: None
                },
                Ecmsc.LIMITS_PRICE.value: {
                    Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                    Ecmsc.LIMITS_PRICE_MAX.value: 50
                },
            }
        }
        min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = \
            AbstractTradingModeCreator.get_min_max_amounts(symbol_market)
        assert min_quantity == 0.5
        assert max_quantity == 100
        assert min_cost is None
        assert max_cost is None
        assert min_price == 0.5
        assert max_price == 50

        # missing all values
        min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = \
            AbstractTradingModeCreator.get_min_max_amounts({})
        assert min_quantity is None
        assert max_quantity is None
        assert min_cost is None
        assert max_cost is None
        assert min_price is None
        assert max_price is None

        # missing all values: asign default
        min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = \
            AbstractTradingModeCreator.get_min_max_amounts({}, "xyz")
        assert min_quantity == "xyz"
        assert max_quantity == "xyz"
        assert min_cost == "xyz"
        assert max_cost == "xyz"
        assert min_price == "xyz"
        assert max_price == "xyz"

        # missing values: assign default

        symbol_market = {
            Ecmsc.LIMITS.value: {
                Ecmsc.LIMITS_AMOUNT.value: {
                    Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                    Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
                },
                Ecmsc.LIMITS_COST.value: {
                    Ecmsc.LIMITS_COST_MIN.value: None,
                    Ecmsc.LIMITS_COST_MAX.value: None
                }
            }
        }
        min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = \
            AbstractTradingModeCreator.get_min_max_amounts(symbol_market, "xyz")
        assert min_quantity == 0.5
        assert max_quantity == 100
        assert min_cost == "xyz"    # None is not a valid value => assign default
        assert max_cost == "xyz"    # None is not a valid value => assign default
        assert min_price == "xyz"
        assert max_price == "xyz"



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import sys
import logging
import traceback
from logging.config import fileConfig

from octobot.constants import LOGGING_CONFIG_FILE, LOGS_FOLDER
from octobot_channels.channels.channel import get_chan
from octobot_commons.logging.logging_util import get_logger
from octobot_commons.pretty_printer import PrettyPrinter
from octobot_evaluators.constants import MATRIX_CHANNEL
from octobot_trading.constants import TICKER_CHANNEL, RECENT_TRADES_CHANNEL, ORDER_BOOK_CHANNEL, KLINE_CHANNEL, \
    OHLCV_CHANNEL, BALANCE_CHANNEL, BALANCE_PROFITABILITY_CHANNEL, TRADES_CHANNEL, POSITIONS_CHANNEL, ORDERS_CHANNEL, \
    MARK_PRICE_CHANNEL, FUNDING_CHANNEL, LIQUIDATIONS_CHANNEL, MINI_TICKER_CHANNEL, ORDER_BOOK_TICKER_CHANNEL
from octobot_trading.channels.exchange_channel import get_chan as get_trading_chan

BOT_CHANNEL_LOGGER = None


def _log_uncaught_exceptions(ex_cls, ex, tb):
    logging.exception(''.join(traceback.format_tb(tb)))
    logging.exception('{0}: {1}'.format(ex_cls, ex))


def init_logger():
    try:
        if not os.path.exists(LOGS_FOLDER):
            os.mkdir(LOGS_FOLDER)

        fileConfig(LOGGING_CONFIG_FILE)
        # overwrite BOT_CHANNEL_LOGGER to apply global logging configuration
        global BOT_CHANNEL_LOGGER
        BOT_CHANNEL_LOGGER = get_logger("OctoBot Channel")
    except KeyError:
        print("Impossible to start OctoBot: the logging configuration can't be found in '" + LOGGING_CONFIG_FILE +
              "' please make sure you are running OctoBot from its root directory.")
        os._exit(-1)

    logger = logging.getLogger("OctoBot Launcher")

    try:
        # Force new log file creation not to log at the previous one's end.
        logger.parent.handlers[1].doRollover()
    except PermissionError:
        print("Impossible to start OctoBot: the logging file is locked, this is probably due to another running "
              "OctoBot instance.")
        os._exit(-1)

    sys.excepthook = _log_uncaught_exceptions
    return logger


async def init_exchange_chan_logger(exchange_id):
    await get_trading_chan(TICKER_CHANNEL, exchange_id).new_consumer(ticker_callback)
    await get_trading_chan(MINI_TICKER_CHANNEL, exchange_id).new_consumer(mini_ticker_callback)
    await get_trading_chan(RECENT_TRADES_CHANNEL, exchange_id).new_consumer(recent_trades_callback)
    await get_trading_chan(ORDER_BOOK_CHANNEL, exchange_id).new_consumer(order_book_callback)
    await get_trading_chan(ORDER_BOOK_TICKER_CHANNEL, exchange_id).new_consumer(order_book_ticker_callback)
    await get_trading_chan(KLINE_CHANNEL, exchange_id).new_consumer(kline_callback)
    await get_trading_chan(OHLCV_CHANNEL, exchange_id).new_consumer(ohlcv_callback)
    await get_trading_chan(BALANCE_CHANNEL, exchange_id).new_consumer(balance_callback)
    await get_trading_chan(BALANCE_PROFITABILITY_CHANNEL, exchange_id).new_consumer(balance_profitability_callback)
    await get_trading_chan(TRADES_CHANNEL, exchange_id).new_consumer(trades_callback)
    await get_trading_chan(LIQUIDATIONS_CHANNEL, exchange_id).new_consumer(liquidations_callback)
    await get_trading_chan(POSITIONS_CHANNEL, exchange_id).new_consumer(positions_callback)
    await get_trading_chan(ORDERS_CHANNEL, exchange_id).new_consumer(orders_callback)
    await get_trading_chan(MARK_PRICE_CHANNEL, exchange_id).new_consumer(mark_price_callback)
    await get_trading_chan(FUNDING_CHANNEL, exchange_id).new_consumer(funding_callback)


async def init_evaluator_chan_logger():
    await get_chan(MATRIX_CHANNEL).new_consumer(matrix_callback)


async def ticker_callback(exchange: str, exchange_id: str, symbol: str, ticker):
    BOT_CHANNEL_LOGGER.debug(f"TICKER : EXCHANGE = {exchange} || SYMBOL = {symbol} || TICKER = {ticker}")


async def mini_ticker_callback(exchange: str, exchange_id: str, symbol: str, mini_ticker):
    BOT_CHANNEL_LOGGER.debug(f"MINI TICKER : EXCHANGE = {exchange} || SYMBOL = {symbol} || MINI TICKER = {mini_ticker}")


async def order_book_callback(exchange: str, exchange_id: str, symbol: str, asks, bids):
    BOT_CHANNEL_LOGGER.debug(
        f"ORDERBOOK : EXCHANGE = {exchange} || SYMBOL = {symbol} || ASKS = {asks} || BIDS = {bids}")


async def order_book_ticker_callback(exchange: str, exchange_id: str, symbol: str,
                                     ask_quantity, ask_price, bid_quantity, bid_price):
    BOT_CHANNEL_LOGGER.debug(
        f"ORDERBOOK TICKER : EXCHANGE = {exchange} || SYMBOL = {symbol} "
        f"|| ASK PRICE / QUANTIY = {ask_price} / {ask_quantity}"
        f"|| BID PRICE / QUANTIY = {bid_price} / {bid_quantity}")


async def ohlcv_callback(exchange: str, exchange_id: str, symbol: str, time_frame, candle):
    BOT_CHANNEL_LOGGER.debug(
        f"OHLCV : EXCHANGE = {exchange} || SYMBOL = {symbol} || TIME FRAME = {time_frame} || CANDLE = {candle}")


async def recent_trades_callback(exchange: str, exchange_id: str, symbol: str, recent_trades):
    BOT_CHANNEL_LOGGER.debug(
        f"RECENT TRADES : EXCHANGE = {exchange} || SYMBOL = {symbol} || RECENT TRADES = {recent_trades}")


async def liquidations_callback(exchange: str, exchange_id: str, symbol: str, liquidations):
    BOT_CHANNEL_LOGGER.debug(
        f"LIQUIDATIONS : EXCHANGE = {exchange} || SYMBOL = {symbol} || LIQUIDATIONS = {liquidations}")


async def kline_callback(exchange: str, exchange_id: str, symbol: str, time_frame, kline):
    BOT_CHANNEL_LOGGER.debug(
        f"KLINE : EXCHANGE = {exchange} || SYMBOL = {symbol} || TIME FRAME = {time_frame} || KLINE = {kline}")


async def mark_price_callback(exchange: str, exchange_id: str, symbol: str, mark_price):
    BOT_CHANNEL_LOGGER.info(f"MARK PRICE : EXCHANGE = {exchange} || SYMBOL = {symbol} || MARK PRICE = {mark_price}")


async def balance_callback(exchange: str, exchange_id: str, balance):
    BOT_CHANNEL_LOGGER.info(f"BALANCE : EXCHANGE = {exchange} || BALANCE = {balance}")


async def balance_profitability_callback(exchange: str, exchange_id: str, profitability, profitability_percent,
                                         market_profitability_percent, initial_portfolio_current_profitability):
    BOT_CHANNEL_LOGGER.info(f"BALANCE PROFITABILITY : EXCHANGE = {exchange} || PROFITABILITY = "
                            f"{PrettyPrinter.portfolio_profitability_pretty_print(profitability, profitability_percent, 'USDT')}")


async def trades_callback(exchange: str, exchange_id: str, symbol: str, trade: dict, old_trade: bool):
    BOT_CHANNEL_LOGGER.info(f"TRADES : EXCHANGE = {exchange} || SYMBOL = {symbol} || TRADE = {trade} "
                            f"|| OLD_TRADE = {old_trade}")


async def orders_callback(exchange: str, exchange_id: str, symbol: str, order: dict, is_closed, is_updated,
                          is_from_bot):
    order_string = f"ORDERS : EXCHANGE = {exchange} || SYMBOL = {symbol} ||"
    if is_closed:
        # order_string += PrettyPrinter.trade_pretty_printer(exchange, order)
        order_string += PrettyPrinter.open_order_pretty_printer(exchange, order)
    else:
        order_string += PrettyPrinter.open_order_pretty_printer(exchange, order)

    order_string += f"|| CLOSED = {is_closed} || UPDATED = {is_updated} || FROM_BOT = {is_from_bot}"
    BOT_CHANNEL_LOGGER.info(order_string)


async def positions_callback(exchange: str, exchange_id: str, symbol: str, position,
                             is_closed, is_updated, is_liquidated, is_from_bot):
    BOT_CHANNEL_LOGGER.info(f"POSITIONS : EXCHANGE = {exchange} || SYMBOL = {symbol} || POSITIONS = {position}"
                            f"|| CLOSED = {is_closed} || UPDATED = {is_updated} || LIQUIDATED = {is_liquidated} "
                            f"|| FROM_BOT = {is_from_bot}")


async def funding_callback(exchange: str, exchange_id: str, symbol: str, funding_rate,
                           next_funding_time, timestamp):
    BOT_CHANNEL_LOGGER.info(f"FUNDING : EXCHANGE = {exchange} || SYMBOL = {symbol} || RATE = {str(funding_rate)}"
                            f"|| NEXT TIME = {str(next_funding_time)} || TIMESTAMP = {str(timestamp)}")


async def matrix_callback(matrix_id,
                          evaluator_name,
                          evaluator_type,
                          eval_note,
                          eval_note_type,
                          exchange_name,
                          cryptocurrency,
                          symbol,
                          time_frame):
    BOT_CHANNEL_LOGGER.debug(f"EXCHANGE = {exchange_name} || EVALUATOR = {evaluator_name} ||"
                             f" CRYPTOCURRENCY = {cryptocurrency} || SYMBOL = {symbol} || TF = {time_frame} || NOTE = {eval_note} [MATRIX id = {matrix_id}] ")



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import pytest

from config import *
from core.octobot import OctoBot
from tests.test_utils.config import load_test_config
from tests.test_utils.bot_management import call_stop_later, start_bot_with_raise

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_run_bot(event_loop):
    # launch a bot
    config = load_test_config()
    config[CONFIG_CRYPTO_CURRENCIES] = {
        "Bitcoin":
            {
                "pairs": ["BTC/USDT"]
            }
    }
    bot = OctoBot(config, ignore_config=True)
    bot.time_frames = [TimeFrames.ONE_MINUTE]
    await bot.initialize()
    await call_stop_later(5, event_loop, bot)
    await start_bot_with_raise(bot)
